@@ -31,6 +31,7 @@
 #' @param opts_BasisFun List of optional configuration parameters passed to the basis function initializer.
 #' @param BasisFunParam Optional list of precomputed basis function parameters.
 #' @param opts Optional list of extra settings passed to inference routines (e.g., \code{stan_iter}, \code{stan_chains}, \code{ndraws}).
+#' @param trend Optional a function that returns the trend of the transformed GP (not to be estimated). If not provided, it defaults to 0.
 #' @param verbose Logical; if \code{TRUE}, print progress and diagnostic messages during computation.
 #'   Defaults to \code{FALSE}.
 #'
@@ -92,8 +93,8 @@ slgp <- function(formula,
                  method,
                  basisFunctionsUsed,
                  interpolateBasisFun = "NN",
-                 nIntegral = 51,
-                 nDiscret = 51,
+                 nIntegral = 101,
+                 nDiscret = 101,
                  hyperparams = NULL,
                  predictorsUpper = NULL,
                  predictorsLower = NULL,
@@ -103,6 +104,7 @@ slgp <- function(formula,
                  opts_BasisFun = list(),
                  BasisFunParam = NULL,
                  opts = list(),
+                 trend = NULL,
                  verbose = FALSE) {
   if(!is.null(seed)){
     set.seed(seed)
@@ -190,6 +192,18 @@ slgp <- function(formula,
   functionValues <- evaluate_basis_functions(parameters=initBasisFun,
                                              X=intermediateQuantities$nodes,
                                              lengthscale=lengthscale)
+  if(is.null(trend)){
+    trend <- function(df){return(rep(0, nrow(df)))}
+    trendValues <- rep(0, nrow(functionValues))
+  }else{
+    dftrend <- as.data.frame(t(t(as.matrix(intermediateQuantities$nodes))*
+                                 c(responseRange[2]-responseRange[1],
+                                   predictorsUpper - predictorsLower)+
+                                 c(responseRange[1], predictorsLower)))
+    colnames(dftrend) <- c(responseName, predictorNames)
+    trendValues <- trend(df=dftrend)
+    rm(dftrend)
+  }
 
   if(sigmaEstimationMethod=="heuristic"){
     Nsim <- 50
@@ -199,6 +213,17 @@ slgp <- function(formula,
     })
     sigma2 <- median(5/resSim)
   }
+
+  if(is.null(opts$discrete)){
+    weightQuadrature <- c(1/nIntegral/2, rep(1/(nIntegral-1), nIntegral-2), 1/nIntegral/2)
+  }else{
+    if(opts$discrete){
+      weightQuadrature <- rep(1/(nIntegral), nIntegral)
+    }else{
+      weightQuadrature <- c(1/nIntegral/2, rep(1/(nIntegral-1), nIntegral-2), 1/nIntegral/2)
+    }
+  }
+
   # Create the data list required for the estimation method selected
   if(interpolateBasisFun == "WNN"){
     if(file.exists("./inst/extdata/composed_model.rds")){
@@ -221,9 +246,10 @@ slgp <- function(formula,
       functionValues = functionValues,
       weightMatrix = temp2,
       indMatrix =temp,
-      weightQuadrature = rep(1/nIntegral, nIntegral),
+      weightQuadrature = weightQuadrature,
       Sigma = diag(sigma2, ncol(functionValues)),
-      mean_x = rep(0, ncol(functionValues))
+      mean_x = rep(0, ncol(functionValues)),
+      trendValues = trendValues
     )
   }else{
 
@@ -242,10 +268,11 @@ slgp <- function(formula,
         p = ncol(functionValues),
         meanFvalues = colMeans(functionValues[intermediateQuantities$indSamplesToNodes,]),
         functionValues = functionValues[!is.na(intermediateQuantities$indNodesToIntegral),],
-        weightQuadrature = rep(1/nIntegral, nIntegral),
         multiplicities=c(as.matrix(table(intermediateQuantities$indSamplesToPredictor))[, 1]),
+        weightQuadrature = weightQuadrature,
         Sigma = diag(sigma2, ncol(functionValues)),
-        mean_x = rep(0, ncol(functionValues))
+        mean_x = rep(0, ncol(functionValues)),
+        trendValues = trendValues
       )
     }
     if(interpolateBasisFun =="NN"){
@@ -256,10 +283,11 @@ slgp <- function(formula,
         p = ncol(functionValues),
         meanFvalues = colMeans(functionValues[intermediateQuantities$indSamplesToNodes,]),
         functionValues = functionValues[!is.na(intermediateQuantities$indNodesToIntegral),],
-        weightQuadrature = rep(1/nIntegral, nIntegral),
+        weightQuadrature = weightQuadrature,
         multiplicities=c(as.matrix(table(intermediateQuantities$indNodesToIntegral[intermediateQuantities$indSamplesToNodes]))[, 1]),
         Sigma = diag(sigma2, ncol(functionValues)),
-        mean_x = rep(0, ncol(functionValues))
+        mean_x = rep(0, ncol(functionValues)),
+        trendValues = trendValues
       )
     }
   }
@@ -335,7 +363,11 @@ slgp <- function(formula,
     fit <- rstan::optimizing(
       object = stan_model,
       data = stan_data,
-      hessian = FALSE)
+      hessian = FALSE,
+      iter = 5000,
+      tol_grad = 1e-12,
+      tol_param = 1e-12,
+      tol_obj = 1e-14)
     # The MAP estimates
     epsilon <- matrix(fit$par, nrow=1)
     # The log-posterior value
@@ -350,6 +382,7 @@ slgp <- function(formula,
               data = data,
               responseName = responseName,
               covariateName = predictorNames,
+              trend=trend,
               method = method,
               predictorsRange = list(upper=predictorsUpper, lower = predictorsLower),
               responseRange = responseRange,
@@ -391,6 +424,7 @@ slgp <- function(formula,
 #' @param seed Optional integer to set the random seed for reproducibility.
 #' @param opts Optional list of additional options passed to inference routines:
 #'   \code{stan_chains}, \code{stan_iter}, \code{ndraws}, etc.
+#' @param trend Optional a function that returns the trend of the transformed GP (not to be estimated). If not provided, it defaults to 0.
 #' @param verbose Logical; if \code{TRUE}, print progress and diagnostic messages during computation.
 #'   Defaults to \code{FALSE}.
 #'
@@ -436,12 +470,13 @@ retrainSLGP <- function(SLGPmodel,
                         epsilonStart =NULL,
                         method,
                         interpolateBasisFun="WNN",
-                        nIntegral=51,
-                        nDiscret=51,
+                        nIntegral=101,
+                        nDiscret=101,
                         hyperparams = NULL,
                         sigmaEstimationMethod = "none",
                         seed=NULL,
                         opts = list(),
+                        trend=NULL,
                         verbose = FALSE) {
   if(!is.null(seed)){
     set.seed(seed)
@@ -500,6 +535,38 @@ retrainSLGP <- function(SLGPmodel,
   functionValues <- evaluate_basis_functions(parameters=initBasisFun,
                                              X=intermediateQuantities$nodes,
                                              lengthscale=lengthscale)
+  if(sigmaEstimationMethod=="heuristic"){
+    Nsim <- 50
+    resSim <- sapply(seq(Nsim), function(i){
+      eps <- rnorm(ncol(functionValues))
+      return(diff(range(functionValues%*%eps)))
+    })
+    sigma2 <- median(5/resSim)
+  }
+  if(is.null(trend)){
+    trend <- SLGPmodel@trend
+  }
+  if(is.null(trend)){
+    trend <- function(df){return(rep(0, nrow(df)))}
+    trendValues <- rep(0, nrow(functionValues))
+  }else{
+    dftrend <- as.data.frame(t(t(as.matrix(intermediateQuantities$nodes))*
+                                 c(responseRange[2]-responseRange[1],
+                                   predictorsUpper - predictorsLower)+
+                                 c(responseRange[1], predictorsLower)))
+    colnames(dftrend) <- c(responseName, predictorNames)
+    trendValues <- trend(df=dftrend)
+    rm(dftrend)
+  }
+  if(is.null(opts$discrete)){
+    weightQuadrature <- c(1/nIntegral/2, rep(1/(nIntegral-1), nIntegral-2), 1/nIntegral/2)
+  }else{
+    if(opts$discrete){
+      weightQuadrature <- rep(1/(nIntegral), nIntegral)
+    }else{
+      weightQuadrature <- c(1/nIntegral/2, rep(1/(nIntegral-1), nIntegral-2), 1/nIntegral/2)
+    }
+  }
 
   # Create the data list required for the estimation method selected
   if(interpolateBasisFun == "WNN"){
@@ -518,9 +585,10 @@ retrainSLGP <- function(SLGPmodel,
       functionValues = functionValues,
       weightMatrix = temp2,
       indMatrix =temp,
-      weightQuadrature = rep(1/nIntegral, nIntegral),
+      weightQuadrature = weightQuadrature,
       Sigma = diag(sigma2, ncol(functionValues)),
-      mean_x = rep(0, ncol(functionValues))
+      mean_x = rep(0, ncol(functionValues)),
+      trendValues = trendValues
     )
   }else{
     stan_model <- stanmodels$likelihoodSimple
@@ -533,10 +601,11 @@ retrainSLGP <- function(SLGPmodel,
         p = ncol(functionValues),
         meanFvalues = colMeans(functionValues[intermediateQuantities$indSamplesToNodes,]),
         functionValues = functionValues[!is.na(intermediateQuantities$indNodesToIntegral),],
-        weightQuadrature = rep(1/nIntegral, nIntegral),
+        weightQuadrature = weightQuadrature,
         multiplicities=c(as.matrix(table(intermediateQuantities$indSamplesToPredictor))[, 1]),
         Sigma = diag(sigma2, ncol(functionValues)),
-        mean_x = rep(0, ncol(functionValues))
+        mean_x = rep(0, ncol(functionValues)),
+        trendValues = trendValues[!is.na(intermediateQuantities$indNodesToIntegral)]
       )
     }
     if(interpolateBasisFun =="NN"){
@@ -547,10 +616,11 @@ retrainSLGP <- function(SLGPmodel,
         p = ncol(functionValues),
         meanFvalues = colMeans(functionValues[intermediateQuantities$indSamplesToNodes,]),
         functionValues = functionValues[!is.na(intermediateQuantities$indNodesToIntegral),],
-        weightQuadrature = rep(1/nIntegral, nIntegral),
+        weightQuadrature = weightQuadrature,
         multiplicities=c(as.matrix(table(intermediateQuantities$indNodesToIntegral[intermediateQuantities$indSamplesToNodes]))[, 1]),
         Sigma = diag(sigma2, ncol(functionValues)),
-        mean_x = rep(0, ncol(functionValues))
+        mean_x = rep(0, ncol(functionValues)),
+        trendValues = trendValues[!is.na(intermediateQuantities$indNodesToIntegral)]
       )
     }
   }
@@ -628,7 +698,11 @@ retrainSLGP <- function(SLGPmodel,
     fit <- rstan::optimizing(
       object = stan_model,
       data = stan_data,
-      hessian = FALSE)
+      hessian = FALSE,
+      iter = 5000,
+      tol_grad = 1e-12,
+      tol_param = 1e-12,
+      tol_obj = 1e-14)
     # The MAP estimates
     epsilon <- matrix(fit$par, nrow=1)
     logPost <- c(fit$value)
@@ -637,12 +711,12 @@ retrainSLGP <- function(SLGPmodel,
   if(method=="none"){
     epsilon <- matrix(nrow=0, ncol=ncol(functionValues))
     logPost <- NaN
-
   }
   gc()
   SLGPmodel@coefficients <- epsilon
   SLGPmodel@hyperparams <- list(sigma2=sigma2, lengthscale=lengthscale)
   SLGPmodel@method <- method
   SLGPmodel@logPost <- logPost
+  SLGPmodel@trend <- trend
   return(SLGPmodel)
 }
